//
//	Modified Object-oriented Realistic Image Synthesis
//	original coded C language in 1989
//	ported to C++ in 2001
//	(c) Copyright & Programmed 2001,2003 by Nakata H.
//	this code is distributed under GPL
//	$Header: /home/nakata/programs/cg/src/RCS/cube.h,v 1.6 2003/07/31 21:00:56 nakata Exp $
//
#ifndef	__CUBE_DEFINED
#define	__CUBE_DEFINED

#include <math.h>
#include "cpg.h"
#include "vector.h"

class cube
{
public:
  vector	size, location;

  cube( double a = 0.0, double b = 0.0, double c = 0.0,
	double d = 0.0, double e = 0.0, double f = 0.0 )
  {
    size.assign( a, b, c );
    location.assign( d, e, f );
  }

  void	assign( double a = 0.0, double b = 0.0, double c = 0.0,
		double d = 0.0, double e = 0.0, double f = 0.0 )
  {
    size.assign( a, b, c );
    location.assign( d, e, f );
  }

  void shift( vector &vec )
  {
    location += vec;
  }

  int intersect( vector &origin, vector &orient, SIDE side,
		 double *solution )
  {
    double	tmp, target;
    vector	diff, tmpvec;
    int		flag = FALSE;

    diff = origin;
    diff -= location;

    if ( side == INNER ) {
      if ( orient.x != 0 ) {
	target = diff.x + ((diff.x > 0) ? (-size.x) : size.x);
	tmp = target / -orient.x;
	if ( tmp >= 0 ) {
	  tmpvec = orient;
	  tmpvec *= tmp;
	  tmpvec += origin;
	  if ( innerPrimitive( tmpvec, INNER ) ) {
	    flag = TRUE;
	    *solution = tmp;
	  }
	}
      }
      if ( orient.y != 0 ) {
	target = diff.y + ((diff.y > 0) ? (-size.y) : size.y);
	tmp = target / -orient.y;
	if ( (tmp >= 0) && (!flag || (tmp < *solution)) ) {
	  tmpvec = orient;
	  tmpvec *= tmp;
	  tmpvec += origin;
	  if ( innerPrimitive( tmpvec, INNER ) ) {
	    flag = TRUE;
	    *solution = tmp;
	  }
	}
      }
      if ( orient.z != 0 ) {
	target = diff.z + ((diff.z > 0) ? (-size.z) : size.z);
	tmp = target / -orient.z;
	if ( (tmp >= 0) && (!flag || (tmp < *solution)) ) {
	  tmpvec = orient;
	  tmpvec *= tmp;
	  tmpvec += origin;
	  if ( innerPrimitive( tmpvec, INNER ) ) {
	    flag = TRUE;
	    *solution = tmp;
	  }
	}
      }
    } else {
      if ( orient.x != 0 ) {
	target = diff.x + ((orient.x < 0) ? size.x : (-size.x));
	tmp = target / -orient.x;
	if ( tmp >= 0 ) {
	  tmpvec = orient;
	  tmpvec *= tmp;
	  tmpvec += origin;
	  if ( innerPrimitive( tmpvec, INNER ) ) {
	    flag = TRUE;
	    *solution = tmp;
	  }
	}
      }
      if ( orient.y != 0 ) {
	target = diff.y + ((orient.y < 0) ? size.y : (-size.y));
	tmp = target / -orient.y;
	if ( (tmp >= 0) && (!flag || (tmp < *solution)) ) {
	  tmpvec = orient;
	  tmpvec *= tmp;
	  tmpvec += origin;
	  if ( innerPrimitive( tmpvec, INNER ) ) {
	    flag = TRUE;
	    *solution = tmp;
	  }
	}
      }
      if ( orient.z != 0 ) {
	target = diff.z + ((orient.z < 0) ? size.z : (-size.z));
	tmp = target / -orient.z;
	if ( (tmp >= 0) && (!flag || (tmp <*solution)) ) {
	  tmpvec = orient;
	  tmpvec *= tmp;
	  tmpvec += origin;
	  if ( innerPrimitive( tmpvec, INNER ) ) {
	    flag = TRUE;
	    *solution = tmp;
	  }
	}
      }
    }

    return (flag);
  }

  int innerPrimitive( vector &point, SIDE side )
  {
    vector	diff;
    int		flag;

    diff = point;
    diff -= location;

    if ( side == INNER ) {
      if ( (size.x *1.0000001 < fabs(diff.x)) ||
	   (size.y *1.0000001 < fabs(diff.y)) ||
	   (size.z *1.0000001 < fabs(diff.z)) ) {
	flag = FALSE;
      } else {
	flag = TRUE;
      }
    } else {
      if ( fabs(diff.x) < size.x * 0.999999 &&
	   fabs(diff.y) < size.y * 0.999999 &&
	   fabs(diff.z) < size.z * 0.999999 ) {
	flag = FALSE;
      } else {
	flag = TRUE;
      }
    }
    return ( flag );
  }

  vector * normalVector( vector &point, SIDE side )
  {
    static vector	vec;
    int			reverse;

    vec = point;
    vec -= location;

    reverse = (side == INNER) ? FALSE : TRUE;
    if ( size.x * 0.999999 < fabs(vec.x) &&
	 fabs(vec.x) < size.x * 1.0000001 ) {
      vec.assign( 1.0, 0, 0 );
      if ( vec.x < 0.0 ) {
	reverse = !reverse;
      }
    } else if ( size.y * 0.999999 < fabs(vec.y) &&
		fabs(vec.y) < size.y * 1.0000001 ) {
      vec.assign( 0, 1.0, 0 );
      if ( vec.y < 0.0 ) {
	reverse = !reverse;
      }
    } else if ( size.z * 0.999999 < fabs(vec.z) &&
		fabs(vec.z) < size.z * 1.0000001 ) {
      vec.assign( 0, 0, 1.0 );
      if ( vec.z < 0.0 ) {
	reverse = !reverse;
      }
    }

    if ( reverse ) {
      vec.x = -vec.x;
      vec.y = -vec.y;
      vec.z = -vec.z;
    }
    return (&vec);
  }

};
#endif	//	__CUBE_DEFINED
// End Of File	---------------------------------------------------------------
