//
//	Modified Object-oriented Realistic Image Synthesis
//	original coded C language in 1989
//	ported to C++ in 2001
//	(c) Copyright & Programmed 2001,2003 by Nakata H.
//	this code is distributed under GPL
//	$Header: /home/nakata/programs/cg/src/RCS/primitive.h,v 1.9 2003/07/31 21:00:56 nakata Exp $
//
#ifndef __PRIMITIVE_DEFINED
#define __PRIMITIVE_DEFINED

#include <math.h>
#include "quadratic3D.h"
#include "quadratic2D.h"
#include "cube.h"
#include "rectangle.h"
typedef enum {NOTHING, CHECKER, STRIPE, CIRCLE, SPOTS }	TEXTURE;

class primitive
{
public:
  union	{	quadratic3D	*p_quad;
    		cube		*p_cube;
	} form3D;
  union	{	quadratic2D	*p_quad2D;
    		rectangle	*p_rect;
  } form2D;			// CPG parameter
  double	areay[2], areax[2]; // CPG parameter
  int		isyarea, isxarea; // CPG parameter
  TEXTURE	texture;
  double	diffuse, reflection, refraction, highlight;
  SIDE		side;
  color		col;
  int		cube_flag;

  int intersect( vector &origin, vector &orient, double *solution )
  {
    if ( cube_flag ) {
      return ( form3D.p_cube->intersect( origin, orient, side, solution ) );
    } else {
      return ( form3D.p_quad->intersect( origin, orient, side, solution ) );
    }
  }

  int innerPrimitive( vector &point )
  {
    if ( cube_flag ) {
      return ( form3D.p_cube->innerPrimitive( point, side ) );
    } else {
      return ( form3D.p_quad->innerPrimitive( point, side ) );
    }
  }

  vector *normalVector( vector &point )
  {
    if ( cube_flag ) {
      return ( form3D.p_cube->normalVector( point, side ) );
    } else {
      return ( form3D.p_quad->normalVector( point, side ) );
    }
  }

  color *getcolor( vector &point )
  {
    static color	col2;
    double		tmp[6];
    int			tmpint, flag;

    col2 = col;
    switch ( texture ) {
    case NOTHING:
      break;
    case CHECKER:
      flag = FALSE;
      tmpint = ((int)point.x) % 20;
      if ( tmpint > 10 || (-10 < tmpint && tmpint <= 0) ) {
	flag = !flag;
      }
      tmpint = ((int)point.y) % 20;
      if ( tmpint > 10 || (-10 < tmpint && tmpint <= 0) ) {
	flag = !flag;
      }
      tmpint = ((int)point.z) % 20;
      if ( tmpint > 10 || (-10 < tmpint && tmpint <= 0) ) {
	flag = !flag;
      }
      if ( flag ) {
	col2.green = 0.0;
      }
      break;
    case STRIPE:
      tmp[0] = sin( point.y / 4 );
      tmp[0] = tmp[0] * tmp[0];
      col2.red = tmp[0];
      tmp[0] = cos( point.y / 4 );
      tmp[0] = tmp[0] * tmp[0];
      col2.green = tmp[0];
      break;
    case CIRCLE:
      tmp[0] = sqrt( point.x * point.x + point.z * point.z ) / 10.0;
      tmp[2] = M_PI * (tmp[0] - floor(tmp[0]));
      tmp[0] = cos(tmp[2]);
      tmp[1] = sin(tmp[2]);
      col2.green = tmp[0] * tmp[0];
      col2.blue = tmp[1] * tmp[1];
      break;
    case SPOTS:
      if ( cube_flag ) {
	tmp[0] = form3D.p_cube->size.x;
	tmp[1] = form3D.p_cube->size.y;
	tmp[2] = form3D.p_cube->size.z;
      } else {
	tmp[0] = form3D.p_quad->a;
	tmp[1] = form3D.p_quad->b;
	tmp[2] = form3D.p_quad->c;
      }
      tmp[0] = point.x * sqrt(tmp[0]);
      tmp[1] = point.y * sqrt(tmp[1]);
      tmp[2] = point.z * sqrt(tmp[2]);
      tmp[3] = sqrt(tmp[0] * tmp[0] + tmp[2] * tmp[2]);
      tmp[4] = ( fabs(tmp[0]) > 0.0001 ) ? fabs(tmp[2]/tmp[0]) : 1e15;
      tmp[5] = ( fabs(tmp[3]) > 0.0001 ) ? fabs(tmp[1]/tmp[3]) : 1e15;
      tmp[4] = atan( tmp[4] ) * 30.0 / M_PI;
      tmp[5] = atan( tmp[5] ) * 30.0 / M_PI;
      tmp[4] -= floor( tmp[4] );
      tmp[5] -= floor( tmp[5] );
      tmp[4] = 0.15 - (0.5 - tmp[4]) * (0.5 - tmp[4]) -
	(0.5 - tmp[5]) * (0.5 - tmp[5]);
      if ( tmp[4] < 0.0 ) {
	tmp[4] = 0.0;
      }
      col2.blue = tmp[4] / 0.3;
      break;
    }
    return (&col2);
  }
};
#endif	//	__PRIMITIVE_DEFINED
// End Of File	---------------------------------------------------------------
