//
//	Modified Object-oriented Realistic Image Synthesis
//	original coded C language in 1989
//	ported to C++ in 2001
//	(c) Copyright & Programmed 2001,2003 by Nakata H.
//	this code is distributed under GPL
//	$Header: /home/nakata/programs/cg/src/RCS/quadratic2D.h,v 1.7 2003/07/31 21:00:56 nakata Exp $
//
#ifndef	__quadratic2D_DEFINED
#define	__quadratic2D_DEFINED

#include "cpg.h"

class quadratic2D
{
public:
  double	a, b, c, d, e, f;

  quadratic2D( double aa = 0.0, double bb = 0.0, double cc = 0.0,
		     double dd = 0.0, double ee = 0.0, double ff = 0.0 )
  {
    a = aa;
    b = bb;
    c = cc;
    d = dd;
    e = ee;
    f = ff;
  }

  void	getYedge( double *p_result )
  {
    double		alpha, beta, gamma, tmp;

    alpha = c * c - 4.0 * a * b;
    beta = 2.0 * c * d - 4.0 * a * e;
    gamma = d * d - 4.0 * a * f;

    if ( (tmp = beta * beta - 4.0 * alpha * gamma) <= 0.0 || alpha == 0.0 ) {
      p_result[0] = p_result[1] = -MAXDOUBLE;
      return;
    }

    tmp = sqrt( tmp );
    if ( alpha > 0.0 ) {
      p_result[0] = (- beta + tmp) / (2.0 * alpha);
      p_result[1] = (- beta - tmp) / (2.0 * alpha);
    } else {
      p_result[0] = (- beta - tmp) / (2.0 * alpha);
      p_result[1] = (- beta + tmp) / (2.0 * alpha);
    }
  }

  int	getXedge( double y, double *p_result )
  {
    double		alpha, beta, gamma, tmp;

    alpha = a;
    beta = c * y + d;
    gamma = b * y * y + e * y + f;

    if ( alpha == 0.0 ) {
      if ( beta == 0.0 ) {
	p_result[0] = p_result[1] = MAXDOUBLE;
	return (FALSE);
      } else {
	p_result[0] = gamma / beta;
	p_result[1] = MAXDOUBLE;
	return (TRUE);
      }
    }

    if ( (tmp = beta * beta - 4.0 * alpha * gamma) <= 0.0 ) {
      p_result[0] = p_result[1] = MAXDOUBLE;
      return (FALSE);
    }

    tmp = sqrt( tmp );
    if ( alpha > 0.0 ) {
      p_result[0] = (- beta - tmp) / (2.0 * alpha);
      p_result[1] = (- beta + tmp) / (2.0 * alpha);
    } else {
      p_result[0] = (- beta + tmp) / (2.0 * alpha);
      p_result[1] = (- beta - tmp) / (2.0 * alpha);
    }
    return (TRUE);
  }
};
#endif	// __quadratic2D_DEFINED
// End Of File	---------------------------------------------------------------
