//
//	Modified Object-oriented Realistic Image Synthesis
//	original coded C language in 1989
//	ported to C++ in 2001
//	(c) Copyright & Programmed 2001,2003 by Nakata H.
//	this code is distributed under GPL
//	$Header: /home/nakata/programs/cg/src/RCS/quadratic3D.h,v 1.9 2003/07/31 21:00:56 nakata Exp $
//
#ifndef	__quadratic3D_DEFINED
#define	__quadratic3D_DEFINED

#include <math.h>
#include "quadratic2D.h"
#include "vector.h"
#include "cpg.h"

class quadratic3D
{
public:
  double	a, b, c, d, e, f, g, h, i, j;

  quadratic3D( double aa = 0.0, double bb = 0.0, double cc = 0.0,
	       double dd = 0.0, double ee = 0.0, double ff = 0.0,
	       double gg = 0.0, double hh = 0.0, double ii = 0.0,
	       double jj = 0.0 )
  {
    a = aa;
    b = bb;
    c = cc;
    d = dd;
    e = ee;
    f = ff;
    g = gg;
    h = hh;
    i = ii;
    j = jj;
  }

  void	assign( double aa = 0.0, double bb = 0.0, double cc = 0.0,
		double dd = 0.0, double ee = 0.0, double ff = 0.0,
		double gg = 0.0, double hh = 0.0, double ii = 0.0,
		double jj = 0.0 )
  {
    a = aa;
    b = bb;
    c = cc;
    d = dd;
    e = ee;
    f = ff;
    g = gg;
    h = hh;
    i = ii;
    j = jj;
  }

  void shift( vector &vec )
  {
    double	gg, hh, ii;
    gg = g;
    hh = h;
    ii = i;
    g += -2.0 * a * vec.x - d * vec.y - e * vec.z;
    h += -2.0 * b * vec.y - d * vec.x - f * vec.z;
    i += -2.0 * c * vec.z - e * vec.x - f * vec.y;
    j += a * vec.x * vec.x + b * vec.y * vec.y + c * vec.z * vec.z
      + d * vec.x * vec.y + e * vec.x * vec.z + f * vec.y * vec.z
      - gg * vec.x - hh * vec.y - ii * vec.z;
  }

  void rotate( vector &rot )
  {
    vector	*cosine, *sine;
    double	tmp[9];
    double	aa, bb, cc, dd, ee, ff, gg, hh, ii;

    cosine = new vector( cos(rot.x), cos(rot.y), cos(rot.z) );
    sine = new vector( sin(rot.x), sin(rot.y), sin(rot.z) );
    tmp[0] = cosine->y * cosine->z;
    tmp[1] = sine->x * sine->y * cosine->z - cosine->x * sine->z;
    tmp[2] = cosine->x * sine->y * cosine->z + sine->x * sine->z;
    tmp[3] = cosine->y * sine->z;
    tmp[4] = sine->x * sine->y * sine->z + cosine->x * cosine->z;
    tmp[5] = cosine->x * sine->y * sine->z - sine->x * cosine->z;
    tmp[6] = -sine->y;
    tmp[7] = sine->x * cosine->y;
    tmp[8] = cosine->x * cosine->y;

    aa = a;
    bb = b;
    cc = c;
    dd = d;
    ee = e;
    ff = f;
    gg = g;
    hh = h;
    ii = i;

    a = aa * tmp[0] * tmp[0] + bb * tmp[3] * tmp[3] + cc * tmp[6] * tmp[6]
      + dd * tmp[0] * tmp[3] + ee * tmp[0] * tmp[6] + ff * tmp[3] * tmp[6];
    b = aa * tmp[1] * tmp[1] + bb * tmp[4] * tmp[4] + cc * tmp[7] * tmp[7]
      + dd * tmp[1] * tmp[4] + ee * tmp[1] * tmp[7] + ff * tmp[4] * tmp[7];
    c = aa * tmp[2] * tmp[2] + bb * tmp[5] * tmp[5] + cc * tmp[8] * tmp[8]
      + dd * tmp[2] * tmp[5] + ee * tmp[2] * tmp[8] + ff * tmp[5] * tmp[8];
    d = 2.0 * (aa * tmp[0] * tmp[1] + bb * tmp[3] * tmp[4]
	       + cc * tmp[6] * tmp[7])
      + dd * (tmp[0] * tmp[4] + tmp[1] * tmp[3])
      + ee * (tmp[0] * tmp[7] + tmp[1] * tmp[6])
      + ff * (tmp[3] * tmp[7] + tmp[4] * tmp[6]);
    e = 2.0 * (aa * tmp[0] * tmp[2] + bb * tmp[3] * tmp[5]
	       + cc * tmp[6] * tmp[8])
      + dd * (tmp[0] * tmp[5] + tmp[2] * tmp[3])
      + ee * (tmp[0] * tmp[8] + tmp[2] * tmp[6])
      + ff * (tmp[3] * tmp[8] + tmp[5] * tmp[6]);
    f = 2.0 * (aa * tmp[1] * tmp[2] + bb * tmp[4] * tmp[5]
	       + cc * tmp[7] * tmp[8])
      + dd * (tmp[1] * tmp[5] + tmp[2] * tmp[4])
      + ee * (tmp[1] * tmp[8] + tmp[2] * tmp[7])
      + ff * (tmp[4] * tmp[8] + tmp[5] * tmp[7]);
    g = gg * tmp[0] + hh * tmp[3] + ii * tmp[6];
    h = gg * tmp[1] + hh * tmp[4] + ii * tmp[7];
    i = gg * tmp[2] + hh * tmp[5] + ii * tmp[8];

    delete( cosine );
    delete( sine );
  }

  int intersect( vector &origin, vector &orient, SIDE side,
		 double *solution )
  {
    double	alpha, beta, gamma;
    double	tmp1, tmp2, tmp3;

    alpha = a * orient.x * orient.x + b * orient.y * orient.y +
      c * orient.z * orient.z + d * orient.x * orient.y +
      e * orient.x * orient.z + f * orient.y * orient.z;
    beta = 2.0 * (a * orient.x * origin.x + b * orient.y * origin.y
		  + c * orient.z * origin.z )
      + d * (orient.x * origin.y + orient.y * origin.x)
      + e * (orient.x * origin.z + orient.z * origin.x)
      + f * (orient.y * origin.z + orient.z * origin.y)
      + g * orient.x + h * orient.y + i * orient.z;
    gamma = a * origin.x * origin.x + b * origin.y * origin.y
      + c * origin.z * origin.z + d * origin.x * origin.y
      + e * origin.x * origin.z + f * origin.y * origin.z
      + g * origin.x + h * origin.y + i * origin.z + j;
    if ( (tmp1 = beta * beta - 4.0 * alpha * gamma) < 0.0 ) {
      return (FALSE);
    }
    if ( alpha != 0.0 ) {	// quadratic surface
      tmp1 = sqrt(tmp1);
      tmp2 = (- beta - tmp1) / (2 * alpha);
      tmp3 = (- beta + tmp1) / (2 * alpha);
      if ( side == OUTER ) {
	*solution = macro_max( tmp2, tmp3 );
      } else {
	*solution = macro_min( tmp2, tmp3 );
      }
      return (TRUE);
    } else {			// plane
      tmp1 = g * orient.x + h * orient.y + i * orient.z;
      if ( tmp1 == 0.0 ) {
	return (FALSE);
      }
      tmp2 = - g * origin.x - h * origin.y - i * origin.z - j;
      *solution = tmp2 / tmp1;
      return (TRUE);
    }
  }

  int innerPrimitive( vector &point, SIDE side )
  {
    double	value;

    value = a * point.x * point.x + b * point.y * point.y
      + c * point.z * point.z + d * point.x * point.y + e * point.x * point.z
      + f * point.y * point.z + g * point.x + h * point.y + i * point.z + j;
    if ( side == OUTER ) {
      value = -value;
    }
    return ( (value > 0) ? FALSE : TRUE );
  }

  vector * normalVector( vector &point, SIDE side )
  {
    static vector	normalV;

    normalV.x = 2.0 * a * point.x + d * point.y + e * point.z + g;
    normalV.y = 2.0 * b * point.y + d * point.x + f * point.z + h;
    normalV.z = 2.0 * c * point.z + e * point.x + f * point.y + i;

    normalV.normalize();

    if ( side == OUTER ) {
      normalV.x = -normalV.x;
      normalV.y = -normalV.y;
      normalV.z = -normalV.z;
    }

    return ( &normalV );
  }

  quadratic2D	*get2Dform( void )
  {
    return (new quadratic2D( g * g - 4.0 * a * j,
			     h * h - 4.0 * b * j,
			     2.0 * g * h - 4.0 * d * j,
			     2.0 * g * i - 4.0 * e * j,
			     2.0 * h * i - 4.0 * f * j,
			     i * i - 4.0 * c * j ) );
  }

  int	isplane( void )
  {
    if ( a == 0.0 && b == 0.0 && c == 0.0 &&
	 d == 0.0 && e == 0.0 && f == 0.0 ) {
      return (TRUE);
    } else {
      return (FALSE);
    }
  }

};
#endif	//	__quadratic3D_DEFINED
// End Of File	---------------------------------------------------------------
