//
//	Modified Object-oriented Realistic Image Synthesis
//	original coded C language in 1989
//	ported to C++ in 2001
//	(c) Copyright & Programmed 2001,2003,2005 by NAKATA H.
//		webmaster@nakata-jp.org
//	this code is distributed under GPL
//

#include "vector.h"
#include "color.h"
#include "pixel.h"
#include "quadratic3D.h"
#include "primitive.h"
#include "light.h"
#include "camera.h"
#include "screen.h"
#include "andor.h"
#include "cpg.h"

#include <math.h>
#include <sstream>

extern camera		cam;
extern screen		scrn;
extern light		light0;
extern double		highlight;
extern int		prim_count, and_count, or_count;
extern primitive	primitives[MAXPRIMITIVES];
extern orclass		or_array[MAXPRIMITIVES];
extern vector		cam_rot;
#ifdef	DEBUG
int	display;
#endif // DEBUG

static void	getYarea( void );
static void	getXarea( double tmpy );
static void	checkCubeArea( primitive *p_prim );
static void	cpgsub( vector &origin, vector &orient, double energy,
			color *p_result );
static int	cpgintersection( vector &origin, vector &orient, 
				 primitive **pp_target, double *p_depth );
static void	tracing( vector &origin, vector &orient, primitive *myself,
		 int level, double energy, color *p_result );
static int	intersection( vector &origin, vector &orient,
			      primitive *myself, primitive **pp_target,
			      double *p_depth );
static double	shadowdetect( vector &origin, vector &orient,
			      primitive *p_myself );

static char	rcsstring[] ="$Header: /home/nakata/programs/cg/src/RCS/raytrace.cpp,v 1.16 2005/04/09 13:18:06 nakata Exp $";

void	cpg( void )
{
  char		c;
  int		x, y, i;
  primitive	*p_prim;
  color		tmpcolor;
  vector	tmpvec[2];
  double	tmpx;
#ifdef	DEBUG
  int		deltax = scrn.width / 80, deltay = scrn.height / 24;
#endif // DEBUG

  getYarea();

  for ( y = 0; y < scrn.height; y++ ) {
    tmpvec[0] = cam.vertical;
    tmpvec[0] *= ((double)(y - scrn.height/2) / (double)(scrn.height/2));
    tmpvec[0] += cam.orient;

    getXarea( (SCREENWIDTH * (double)scrn.height / (double)scrn.width) *
	      ((double)(scrn.height/2 - y) / (double)(scrn.height/2)) /
	      SCREENDEPTH );

    c = '.';
    for ( x = 0; x < scrn.width; x++ ) {
      tmpvec[1] = cam.horizontal;
      tmpvec[1] *= ((double)(x - scrn.width/2) / (double)(scrn.width/2));
      tmpvec[1] += tmpvec[0];

#ifdef	DEBUG
      if ( (y % deltay == 0) && (x % deltax == 0) ) {
	display = TRUE;
      } else {
	display = FALSE;
      }
#endif // DEBUG

      tmpx = SCREENWIDTH * ((double)(x - scrn.width/2) /
			    (double)(scrn.width/2)) / SCREENDEPTH;
      for ( p_prim = primitives, i = 0; i < prim_count; i++, p_prim++ ) {
	while ( p_prim->areax[0] < tmpx ) {
	  p_prim->areax[0] = p_prim->areax[1];
	  p_prim->areax[1] = MAXDOUBLE;
	  p_prim->isxarea = !p_prim->isxarea;
	  c = 'x';
	}
      }

      cpgsub( cam.point, tmpvec[1], 1.0, &tmpcolor );
      scrn.col[x + scrn.width * y] = tmpcolor;

#ifdef	DEBUG
      if ( display ) {
	std::cout << c;
	c = '.';
      }
#endif // DEBUG

    }

#ifdef	DEBUG
    if ( y % deltay == 0 ) {
      std::cout << '\n';
    }
#endif // DEBUG
  }
  for ( i = 0; i < prim_count; i++ ) {
    delete primitives[i].form2D.p_quad2D;
  }
}

static void	getYarea( void )
{
  vector	tmpvec[2];
  double	tmpy;
  int		i;
  primitive	*p_prim;
  quadratic3D	quad3D;

  tmpvec[0] = cam.point;
  tmpvec[0] *= -1.0;
  tmpvec[1] = cam_rot;
  tmpvec[1] *= -1.0;
  tmpy = (SCREENWIDTH * (double)scrn.height / (double)scrn.width) *
    2.0 / SCREENDEPTH;		// tmpy must be above of screen upper line
  for ( i = 0; i < prim_count; i++ ) {
    p_prim = &primitives[i];
    if ( p_prim->side == OUTER ||
	 (!p_prim->cube_flag && p_prim->form3D.p_quad->isplane()) ) {
      p_prim->areay[0] = -MAXDOUBLE;
      p_prim->areay[1] = -MAXDOUBLE;
      p_prim->isyarea = TRUE;
      continue;
    }

    if ( p_prim->cube_flag ) {
      checkCubeArea( p_prim );
      p_prim->isyarea = FALSE;
      p_prim->areay[0] = p_prim->form2D.p_rect->y_max;
      p_prim->areay[1] = p_prim->form2D.p_rect->y_min;
      while ( p_prim->areay[0] > tmpy ) {
	p_prim->areay[0] = p_prim->areay[1];
	p_prim->areay[1] = -MAXDOUBLE;
	p_prim->isyarea = !p_prim->isyarea;
      }
      continue;
    }

    quad3D = *(p_prim->form3D.p_quad);
    quad3D.shift( tmpvec[0] );
    quad3D.rotate( tmpvec[1] );
    p_prim->form2D.p_quad2D = quad3D.get2Dform();
    p_prim->form2D.p_quad2D->getYedge( p_prim->areay );
    while ( p_prim->areay[0] > tmpy ) {
      p_prim->areay[0] = p_prim->areay[1];
      p_prim->areay[1] = -MAXDOUBLE;
    }
    p_prim->isyarea = p_prim->form2D.p_quad2D->getXedge( tmpy, p_prim->areax );
  }
}

static void	getXarea( double tmpy )
{
  double	tmpx, depth;
  int		i;
  vector	tmpvec;
  primitive	*p_prim;

  tmpx = SCREENWIDTH * -2.0 / SCREENDEPTH; // tmpx must be out of left edge
  tmpvec.assign( tmpx, tmpy, 1.0 );
  tmpvec.rotx( cam_rot.x );
  tmpvec.roty( cam_rot.y );
  for ( i = 0; i < prim_count; i++ ) {
    p_prim = &primitives[i];
    if ( p_prim->side == OUTER ) {
      p_prim->areax[0] = MAXDOUBLE;
      p_prim->areax[1] = MAXDOUBLE;
      p_prim->isxarea = TRUE;
      continue;
    } else if ( p_prim->cube_flag ) {
      while ( p_prim->areay[0] > tmpy ) {
	p_prim->areay[0] = p_prim->areay[1];
	p_prim->areay[1] = -MAXDOUBLE;
	p_prim->isyarea = !p_prim->isyarea;
      }
      if ( !p_prim->isyarea ) {
	continue;
      }
      p_prim->isxarea = FALSE;
      p_prim->areax[0] = p_prim->form2D.p_rect->x_min;
      p_prim->areax[1] = p_prim->form2D.p_rect->x_max;
      while ( p_prim->areax[0] < tmpx ) {
	p_prim->areax[0] = p_prim->areax[1];
	p_prim->areax[1] = -MAXDOUBLE;
	p_prim->isxarea = !p_prim->isxarea;
      }
      continue;
    } else if ( p_prim->form3D.p_quad->isplane() ) {
      p_prim->areax[0] = MAXDOUBLE;
      p_prim->areax[1] = MAXDOUBLE;
      p_prim->isxarea = TRUE;
      continue;
    }
    
    while ( p_prim->areay[0] > tmpy ) {
      p_prim->areay[0] = p_prim->areay[1];
      p_prim->areay[1] = -MAXDOUBLE;
      p_prim->isyarea = !p_prim->isyarea;
    }
    if ( !p_prim->isyarea ) {
      p_prim->areax[0] = MAXDOUBLE;
      p_prim->areax[1] = MAXDOUBLE;
      p_prim->isxarea = FALSE;
      continue;
    }
    p_prim->form2D.p_quad2D->getXedge( tmpy, p_prim->areax );
    while ( p_prim->areax[0] < tmpx ) {
      p_prim->areax[0] = p_prim->areax[1];
      p_prim->areax[1] = MAXDOUBLE;
    }
    if ( p_prim->form3D.p_quad->intersect( cam.point, tmpvec, INNER,
					   &depth ) ) {
      p_prim->isxarea = TRUE;
    } else {
      p_prim->isxarea = FALSE;
    }
  }
}

static void	checkCubeArea( primitive *p_prim )
{
  rectangle	*p_rect = new rectangle( MAXDOUBLE, -MAXDOUBLE,
					 MAXDOUBLE, -MAXDOUBLE );
  vector	tmpvec[4];
  vector	vecx, vecy, vecz;

  p_prim->form2D.p_rect = p_rect;
  vecx.assign( p_prim->form3D.p_cube->size.x, 0.0, 0.0 );
  vecx.roty( -cam_rot.y );
  vecx.rotx( -cam_rot.x );
  vecy.assign( 0.0, p_prim->form3D.p_cube->size.y, 0.0 );
  vecy.roty( -cam_rot.y );
  vecy.rotx( -cam_rot.x );
  vecz.assign( 0.0, 0.0, p_prim->form3D.p_cube->size.z );
  vecz.roty( -cam_rot.y );
  vecz.rotx( -cam_rot.x );

  tmpvec[0] = p_prim->form3D.p_cube->location;
  tmpvec[0] -= cam.point;
  tmpvec[0].roty( -cam_rot.y );
  tmpvec[0].rotx( -cam_rot.x );

  tmpvec[1] = tmpvec[0];
  tmpvec[2] = tmpvec[1] += vecx;
  tmpvec[3] = tmpvec[2] += vecy;
  tmpvec[3] += vecz;
  p_rect->stretch( tmpvec[3] );
  tmpvec[3] = tmpvec[2];
  tmpvec[3] -= vecz;
  p_rect->stretch( tmpvec[3] );

  tmpvec[2] = tmpvec[1];
  tmpvec[3] = tmpvec[2] -= vecy;
  tmpvec[3] += vecz;
  p_rect->stretch( tmpvec[3] );
  tmpvec[3] = tmpvec[2];
  tmpvec[3] -= vecz;
  p_rect->stretch( tmpvec[3] );

  tmpvec[1] = tmpvec[0];
  tmpvec[2] = tmpvec[1] -= vecx;
  tmpvec[3] = tmpvec[2] += vecy;
  tmpvec[3] += vecz;
  p_rect->stretch( tmpvec[3] );
  tmpvec[3] = tmpvec[2];
  tmpvec[3] -= vecz;
  p_rect->stretch( tmpvec[3] );

  tmpvec[2] = tmpvec[1];
  tmpvec[3] = tmpvec[2] -= vecy;
  tmpvec[3] += vecz;
  p_rect->stretch( tmpvec[3] );
  tmpvec[3] = tmpvec[2];
  tmpvec[3] -= vecz;
  p_rect->stretch( tmpvec[3] );
}

static void	cpgsub( vector &origin, vector &orient, double energy,
			color *p_result )
{
  primitive	*p_prim;
  double	depth, energy2, brightness, highlight2, d_tmp;
  vector	on_surface, to_surface, to_light, normal, vec_reflect;

  if ( energy < TRACELIMIT ) {
    return;
  }

  if ( !cpgintersection( origin, orient, &p_prim, &depth ) ) {
    p_result->assign( 0.0, 0.0, 0.0 );
    return;			// this return command need not
  } else {
    highlight2 = 0;
    on_surface = orient;
    on_surface *= depth;
    on_surface += origin;
    normal = *(p_prim->normalVector(on_surface));
    to_light.assign( -light0.x, -light0.y, -light0.z );
    to_surface = orient;
    to_surface.normalize();
    d_tmp = -2.0 * to_surface.naiseki(normal);
    vec_reflect = normal;
    vec_reflect *= d_tmp;
    vec_reflect += to_surface;
    if ( normal.naiseki(to_light) < 0.0 ||
	 (energy2 = shadowdetect( on_surface, to_light, p_prim )) < 0.01 ) {
      brightness = BACKLIGHT;
    } else {
      brightness = - normal.x * light0.x - normal.y * light0.y -
	normal.z * light0.z;
      if ( p_prim->highlight > 0.0 && highlight > 0.0 ) {
	highlight2 = vec_reflect.naiseki(to_light);
	if ( highlight2 <= 0.0 ) {
	  highlight2 = 0.0;
	} else {
	  highlight2 *= highlight2;
	  highlight2 *= highlight2;
	  highlight2 *= brightness * p_prim->highlight * highlight;
	}
      }
      if ( brightness < BACKLIGHT ) {
	energy2 = 1.0;
	brightness = BACKLIGHT;
      }
    }

    *p_result = *(p_prim->getcolor(on_surface));
    *p_result *= brightness * p_prim->diffuse;
    *p_result += highlight2;
    *p_result *= energy;
    //    *p_result *= energy2;

    if ( p_prim->reflection > 0.0 ) {
      color	col;

      tracing( on_surface, vec_reflect, p_prim, 1, energy * p_prim->reflection,
	       &col );
	*p_result += col;
    }
  }
}

static int	cpgintersection( vector &origin, vector &orient, 
				 primitive **pp_target, double *p_depth )
{
  int		i, j, k, l, flag;
  primitive	*p_prim;
  double	tmpdepth;
  vector	point;
  orclass	*p_or;
  andclass	*p_and;

  flag = FALSE;
  for ( i = 0; i < or_count; i++ ) {
    p_or = &(or_array[i]);
    if ( (p_or->p_bound != NULL) &&
	 (!p_or->p_bound->innerPrimitive( origin )) ) {
      continue;		// out of bounds
    }
    for ( j = 0; j < p_or->length; j++ ) {
      p_and = p_or->p_and[j];

      for ( k = 0; k < p_and->length; k++ ) {
	if ( !(p_and->p_prim[k]->isxarea) ) {
	  goto nextAnd;
	}
      }

      for ( k = 0; k < p_and->length; k++ ) {
	p_prim = p_and->p_prim[k];
	if ( !p_prim->intersect( origin, orient, &tmpdepth )
	     || (tmpdepth <= 0.0) ) {
	  continue;		// no intersection
	}
	if ( flag  && (*p_depth < tmpdepth) ) {
	  continue;		// this primitive is far
	}
	point = orient;
	point *= tmpdepth;
	point += origin;
	for ( l = 0; l < p_and->length; l++ ) {
	  if ( (k != l) && !(p_and->p_prim[l]->innerPrimitive( point )) ) {
	    goto checkNext;
	  }
	}
	flag = TRUE;
	*pp_target = p_prim;
	*p_depth = tmpdepth;
      checkNext:;
      }
    nextAnd:;
    }
  }
  return (flag);
}

void	raytrace( void )
{
  int		x, y;
  color		tmpcolor;
  vector	tmpvec[2];
#ifdef	DEBUG
  int		deltax = scrn.width / 80, deltay = scrn.height / 24;
#endif // DEBUG

  for ( y = 0; y < scrn.height; y++ ) {
    tmpvec[0] = cam.vertical;
    tmpvec[0] *= ((double)(y - scrn.height/2) / (double)(scrn.height/2));
    tmpvec[0] += cam.orient;
    for ( x = 0; x < scrn.width; x++ ) {
      tmpvec[1] = cam.horizontal;
      tmpvec[1] *= ((double)(x - scrn.width/2) / (double)(scrn.width/2));
      tmpvec[1] += tmpvec[0];

#ifdef	DEBUG
      if ( (y % deltay == 0) && (x % deltax == 0) ) {
	display = TRUE;
      } else {
	display = FALSE;
      }
#endif // DEBUG

      tracing( cam.point, tmpvec[1], NULL, 0, 1.0, &tmpcolor );
      scrn.col[x + scrn.width * y] = tmpcolor;

    }

#ifdef	DEBUG
    if ( y % deltay == 0 ) {
      std::cout<<'\n';
    }
#endif // DEBUG
  }
}

static void	tracing( vector &origin, vector &orient, primitive *myself,
			 int level, double energy, color *p_result )
{
  primitive	*p_prim;
  double	depth, energy2, brightness, highlight2, d_tmp;
  vector	on_surface, to_surface, to_light, normal, vec_reflect;
#ifdef	DEBUG
  int		i;
#endif // DEBUG

  if ( level > MAXREFLECT || energy < TRACELIMIT ) {
    return;
  }

  if ( !intersection( origin, orient, myself, &p_prim, &depth ) ) {
    p_result->assign( 0.0, 0.0, 0.0 );
#ifdef	DEBUG
    if ( display && level == 0 ) {
      std::cout<< '.';
    }
#endif // DEBUG
    return;			// this return command need not
  } else {
    highlight2 = 0;
    on_surface = orient;
    on_surface *= depth;
    on_surface += origin;
    normal = *(p_prim->normalVector(on_surface));
    to_light.assign( -light0.x, -light0.y, -light0.z );
    to_surface = orient;
    to_surface.normalize();
    d_tmp = -2.0 * to_surface.naiseki(normal);
    vec_reflect = normal;
    vec_reflect *= d_tmp;
    vec_reflect += to_surface;
    if ( normal.naiseki(to_light) < 0.0 ||
	 (energy2 = shadowdetect( on_surface, to_light, p_prim )) < 0.01 ) {
      brightness = BACKLIGHT;
    } else {
      brightness = - normal.x * light0.x - normal.y * light0.y -
	normal.z * light0.z;
      if ( p_prim->highlight > 0.0 && highlight > 0.0 ) {
	highlight2 = vec_reflect.naiseki(to_light);
	if ( highlight2 <= 0.0 ) {
	  highlight2 = 0.0;
	} else {
	  highlight2 *= highlight2;
	  highlight2 *= highlight2;
	  highlight2 *= brightness * p_prim->highlight * highlight;
	}
      }
      if ( brightness < BACKLIGHT ) {
	energy2 = 1.0;
	brightness = BACKLIGHT;
      }
    }

    *p_result = *(p_prim->getcolor(on_surface));
    *p_result *= brightness * p_prim->diffuse;
    *p_result += highlight2;
    *p_result *= energy;
    //    *p_result *= energy2;

    if ( p_prim->reflection > 0.0 ) {
      color	col;

      tracing( on_surface, vec_reflect, p_prim, level + 1,
	       energy * p_prim->reflection, &col );
	*p_result += col;
    }

#ifdef	DEBUG
    if ( display && level == 0 ) {
      for ( i = 0; i < prim_count; i++ ) {
	if ( &(primitives[i]) == p_prim ) {
	  std::cout<<(char)(i + '0');
	  break;
	}
      }
    }
#endif // DEBUG

  }
}

static int	intersection( vector &origin, vector &orient,
			      primitive *myself, primitive **pp_target,
			      double *p_depth )
{
  int		i, j, k, l, flag;
  primitive	*p_prim;
  double	tmpdepth;
  vector	point;
  orclass	*p_or;
  andclass	*p_and;

  flag = FALSE;
  for ( i = 0; i < or_count; i++ ) {
    p_or = &(or_array[i]);
    if ( (p_or->p_bound != NULL) &&
	 (!p_or->p_bound->innerPrimitive( origin )) &&
	 ((!p_or->p_bound->intersect( origin, orient, &tmpdepth )) ||
	  (tmpdepth < 0.0)) ) {
      continue;		// out of bounds
    }
    for ( j = 0; j < p_or->length; j++ ) {
      p_and = p_or->p_and[j];
      for ( k = 0; k < p_and->length; k++ ) {
	p_prim = p_and->p_prim[k];
	if ( p_prim == myself ) {
	  continue;		// skip intersection of start primitive
	}
	if ( !p_prim->intersect( origin, orient, &tmpdepth )
	     || (tmpdepth <= 0.0) ) {
	  continue;		// no intersection
	}
	if ( flag  && (*p_depth < tmpdepth) ) {
	  continue;		// this primitive is far
	}
	point = orient;
	point *= tmpdepth;
	point += origin;
	for ( l = 0; l < p_and->length; l++ ) {
	  if ( (k != l) && !(p_and->p_prim[l]->innerPrimitive( point )) ) {
	    goto checkNext;
	  }
	}
	flag = TRUE;
	*pp_target = p_prim;
	*p_depth = tmpdepth;
      checkNext:;
      }
    }
  }
  return (flag);
}

static double	shadowdetect( vector &origin, vector &orient,
			      primitive *p_myself )
{
  int		i, j, k, l;
  primitive	*p_prim;
  double	tmpdepth, energy;
  vector	point;
  orclass	*p_or;
  andclass	*p_and;

  energy = 1.0;
  for ( i = 0; i < or_count; i++ ) {
    p_or = &(or_array[i]);
    if ( (p_or->p_bound != NULL) &&
	 (!p_or->p_bound->innerPrimitive( origin )) &&
	 (!p_or->p_bound->intersect( origin, orient, &tmpdepth ) ||
	  (tmpdepth < 0.0)) ) {
      continue;		// out of bounds
    }
    for ( j = 0; j < p_or->length; j++ ) {
      p_and = p_or->p_and[j];
      for ( k = 0; k < p_and->length; k++ ) {
	p_prim = p_and->p_prim[k];
	if ( (p_prim == p_myself) && p_prim->cube_flag ) {
	  continue;
	}
	if ( !p_prim->intersect( origin, orient, &tmpdepth )
	     || (tmpdepth <= 0.0) ) {
	  continue;		// no intersection
	}
	point = orient;
	point *= tmpdepth;
	point += origin;
	for ( l = 0; l < p_and->length; l++ ) {
	  if ( (k != l) && !(p_and->p_prim[l]->innerPrimitive( point )) ) {
	    goto checkNext;
	  }
	}
	return ( 0.0 );
      checkNext:;
      }
    }
  }
  return (energy);
}
// End Of File	---------------------------------------------------------------
