//
//	Modified Object-oriented Realistic Image Synthesis
//	original coded C language in 1989
//	ported to C++ in 2001
//	(c) Copyright & Programmed 2001,2003,2005 by NAKATA H.
//		webmaster@nakata-jp.org
//	this code is distributed under GPL
//	$Header: /home/nakata/programs/cg/src/RCS/screen.h,v 1.9 2005/04/09 13:18:06 nakata Exp $
//
#ifndef __SCREEN_DEFINED
#define __SCREEN_DEFINED

//#include <stdio.h>
#include <cstdio>
#include <iostream>
#include "vector.h"
#include "cpg.h"
#include "pixel.h"

class screen
{
public:
  int	width, height;
  pixel	*col;

  screen( int x = 320, int y = 200 )
  {
    width = x;
    height = y;
    col = new pixel[x*y];
  }

  ~screen()
  {
    delete col;
  }

  void assign( int x = 320, int y = 200 )
  {
    width = x;
    height = y;
    delete col;
    col = new pixel[x*y];
  }

  int	shortout( FILE *fp, short s)
  {
    if ( fputc( s & 0xff, fp ) == EOF ) {
      std::cerr << "fputc() failed.\n";
      return (-1);
    }
    if ( fputc( (s >> 8) & 0xff, fp ) == EOF ) {
      std::cerr << "fputc() failed.\n";
      return (-1);
    }
    return (0);
  }

  int	longout( FILE *fp, long l)
  {
    if ( fputc( l & 0xff, fp ) == EOF ) {
      std::cerr << "fputc() failed.\n";
      return (-1);
    }
    if ( fputc( (l >> 8) & 0xff, fp ) == EOF ) {
      std::cerr << "fputc() failed.\n";
      return (-1);
    }
    if ( fputc( (l >> 16) & 0xff, fp ) == EOF ) {
      std::cerr << "fputc() failed.\n";
      return (-1);
    }
    if ( fputc( (l >> 24) & 0xff, fp ) == EOF ) {
      std::cerr << "fputc() failed.\n";
      return (-1);
    }
    return (0);
  }

  int bmpout( char *filename )
  {
    FILE	*fp;
    char	c;
    short	s;
    long	l;
    int		x, y;
    int		pudlen;

    if ( (fp = fopen(filename,"wb")) == NULL ) {
      std::cerr << "fopen(" << filename << ") failed.\n";
      return (-1);
    }

    if ( fwrite( "BM", 1, 2, fp ) < 2 ) {
      std::cerr << "fwrite() failed.\n";
      return (-1);
    }

    pudlen = (width * 3) % 4;
    pudlen = (4 - pudlen) & 3;
    // file size
    l = (width * 3 + pudlen) * height + 0x36;
    if ( longout(fp, l) < 0 ) {
      return (-1);
    }

    // reserved
    if ( longout(fp, 0) < 0 ) {
      return (-1);
    }

    // offset
    if ( longout(fp, 0x36) < 0 ) {
      return (-1);
    }

    // header size
    if ( longout(fp, 0x28) < 0 ) {
      return (-1);
    }

    // horizontal dots
    if ( longout(fp, width) < 0 ) {
      return (-1);
    }

    // vertical dots
    if ( longout(fp, height) < 0 ) {
      return (-1);
    }

    // plane
    if ( shortout(fp, 1) < 0 ) {
      return (-1);
    }

    // bits / pixel
    if ( shortout(fp, 24) < 0 ) {
      return (-1);
    }

    // compression
    if ( longout(fp, 0) < 0 ) {
      return (-1);
    }

    // bitmap data
    if ( longout(fp, 0) < 0 ) {
      return (-1);
    }

    // horizontal resolution
    if ( longout(fp, 0x6893) < 0 ) {
      return (-1);
    }

    // vertical resolution
    if ( longout(fp, 0x6893) < 0 ) {
      return (-1);
    }

    // colors
    if ( longout(fp, 0) < 0 ) {
      return (-1);
    }

    // important colors
    if ( longout(fp, 0) < 0 ) {
      return (-1);
    }

    for ( y = height - 1; y >= 0; y-- ) {
      for ( x = 0; x < width; x++ ) {
	c = col[x + y * width].blue;
	if ( fputc( c, fp) == EOF ) {
	  std::cerr << "fputc() failed.\n";
	  return (-1);
	}
	c = col[x + y * width].green;
	if ( fputc( c, fp) == EOF ) {
	  std::cerr << "fputc() failed.\n";
	  return (-1);
	}
	c = col[x + y * width].red;
	if ( fputc( c, fp) == EOF ) {
	  std::cerr << "fputc() failed.\n";
	  return (-1);
	}
      }
      for ( x = 0; x < pudlen; x++) {		// adjust to word align
	if ( fputc( 0, fp) == EOF ) {
	  std::cerr << "fputc() failed.\n";
	  return (-1);
	}
      }
    }

    fclose( fp );
    return (0);
  }
};
#endif	//	__SCREEN_DEFINED
// End Of File	---------------------------------------------------------------
