//
//	Modified Object-oriented Realistic Image Synthesis
//	original coded C language in 1989
//	ported to C++ in 2001
//	(c) Copyright & Programmed 2001,2003,2005 by Nakata H.
//		webmaster@nakata-jp.org
//	this code is distributed under GPL
//	$Header: /home/nakata/programs/cg/src/RCS/vector.h,v 1.9 2005/04/09 13:18:06 nakata Exp $
//
#ifndef __VECTER_DEFINED
#define __VECTER_DEFINED

#include <math.h>

class vector
{
public:
  double	x, y, z;

  vector( double xx = 0.0, double yy = 0.0, double zz = 0.0 )
  {
    x = xx;
    y = yy;
    z = zz;
  }

  void	assign( double xx = 0.0, double yy = 0.0, double zz = 0.0 )
  {
    x = xx;
    y = yy;
    z = zz;
  }

  double length (void)
  {
    return( sqrt( x * x + y * y + z * z ) );
  }

  vector normalize (void)
  {
    double len;

    len = length();
    if ( len > 0.0 ) {
      x /= len;
      y /= len;
      z /= len;
    }
    return (*this);
  }

  double naiseki( vector &vec )
  {
    return ( x * vec.x + y * vec.y + z * vec.z );
  }

  void rotx( double theta )
  {
    double	sine, cosine;
    double	yy, zz;

    sine = sin(theta);
    cosine = cos(theta);
    yy = y;
    zz = z;
    y = yy * cosine + zz * sine;
    z = - yy * sine + zz * cosine;
  }

  void roty( double theta )
  {
    double	sine, cosine;
    double	xx, zz;

    sine = sin(theta);
    cosine = cos(theta);
    xx = x;
    zz = z;
    x = xx * cosine - zz * sine;
    z = xx * sine + zz * cosine;
  }

  void rotz( double theta )
  {
    double	sine, cosine;
    double	xx, yy;

    sine = sin(theta);
    cosine = cos(theta);
    xx = x;
    yy = y;
    x = xx * cosine - yy * sine;
    y = xx * sine + yy * cosine;
  }

  vector operator += (vector &parameter)
  {
    x += parameter.x;
    y += parameter.y;
    z += parameter.z;
    return (*this);
  }

  vector operator -= (vector &parameter)
  {
    x -= parameter.x;
    y -= parameter.y;
    z -= parameter.z;
    return (*this);
  }

  vector operator *= (double scale)
  {
    x *= scale;
    y *= scale;
    z *= scale;
    return (*this);
  }

  vector operator /= (double scale)
  {
    x /= scale;
    y /= scale;
    z /= scale;
    return (*this);
  }

  vector operator *= (vector &parameter)
  {
    x *= parameter.x;
    y *= parameter.y;
    z *= parameter.z;
    return (*this);
  }

#ifdef	NOTDEF
  static vector	reflect;

  vector * reflectVector( vector &view, vector &normal )
  {
    double	dcosine;

    dcosine = 2.0 * (- view.x * normal.x - view.y * normal.y
		     - view.z * normal.z);
    reflect.x = view.x + dcosine * normal.x;
    reflect.y = view.y + dcosine * normal.y;
    reflect.z = view.z + dcosine * normal.z;
    return ( &reflect );
    }
#endif /* NOTDEF */

};
#endif	//	__VECTER_DEFINED
// End Of File	---------------------------------------------------------------
